@@ -60,8 +60,9 @@ class FacebookRequestError(FacebookError):
         self._api_error_code = None
         self._api_error_type = None
         self._api_error_message = None
+        self._api_blame_field_specs = None
 
-        if self._body is not None and 'error' in self._body:
+        if self._body and 'error' in self._body:
             self._error = self._body['error']
             if 'message' in self._error:
                 self._api_error_message = self._error['message']
@@ -69,6 +70,9 @@ class FacebookRequestError(FacebookError):
                 self._api_error_code = self._error['code']
             if 'type' in self._error:
                 self._api_error_type = self._error['type']
+            if 'error_data' in self._error:
+                self._api_blame_field_specs = \
+                    self._error['error_data']['blame_field_specs']
         else:
             self._error = None
 
@@ -104,6 +108,9 @@ class FacebookRequestError(FacebookError):
     def api_error_type(self):
         return self._api_error_type
 
+    def api_blame_field_specs(self):
+        return self._api_blame_field_specs
+
 
 class FacebookBadObjectError(FacebookError):
     """Raised when a guarantee about the object validity fails."""
