@@ -27,6 +27,7 @@ from facebookads.exceptions import FacebookBadObjectError
 from facebookads.api import FacebookAdsApi
 from facebookads.mixins import (
     CanArchive,
+    CanValidate,
     CannotCreate,
     CannotDelete,
     CannotUpdate,
@@ -74,14 +75,8 @@ class EdgeIterator(object):
                 is the parameter name and its value is a string or an object
                 which can be JSON-encoded.
         """
-        fields = ','.join(fields or
-                          target_objects_class.get_default_read_fields())
-
-        self._params = {} if params is None else params.copy()
-
-        if fields:
-            self._params['fields'] = fields
-
+        self._params = dict(params or {})
+        target_objects_class._assign_fields_to_params(fields, self._params)
         self._source_object = source_object
         self._target_objects_class = target_objects_class
         self._path = (
@@ -133,7 +128,7 @@ class EdgeIterator(object):
         self._params['summary'] = True
 
         response = self._source_object.get_api_assured().call(
-            FacebookAdsApi.HTTP_METHOD_GET,
+            'GET',
             self._path,
             params=self._params,
         ).json()
@@ -202,7 +197,7 @@ class AbstractObject(collections.MutableMapping):
         return "<%s> %s" % (
             self.__class__.__name__,
             json.dumps(
-                self._data,
+                self.export_value(self._data),
                 sort_keys=True,
                 indent=4,
                 separators=(',', ': ')
@@ -237,6 +232,14 @@ class AbstractObject(collections.MutableMapping):
         """
         cls._default_read_fields = fields
 
+    @classmethod
+    def _assign_fields_to_params(cls, fields, params):
+        """Applies fields to params in a consistent manner."""
+        if fields is None:
+            fields = cls.get_default_read_fields()
+        if fields:
+            params['fields'] = ','.join(fields)
+
     def _set_data(self, data):
         """
         An AbstractObject does not keep history so _set_data is an alias for
@@ -287,7 +290,7 @@ class AbstractCrudObject(AbstractObject):
         super(AbstractCrudObject, self).__init__()
 
         self._changes = {}
-        self[self.Field.id] = fbid
+        self._data[self.Field.id] = fbid
         self._parent_id = parent_id
         self._api = api
 
@@ -303,8 +306,7 @@ class AbstractCrudObject(AbstractObject):
 
     def __delitem__(self, key):
         del self._data[key]
-        if key in self._changes:
-            del self._changes[key]
+        self._changes.pop(key, None)
 
     def __eq__(self, other):
         """Two objects are the same if they have the same fbid."""
@@ -313,7 +315,7 @@ class AbstractCrudObject(AbstractObject):
             isinstance(other, self.__class__) and
 
             # Both have id's
-            self.get_id() is not None and other.get_id() is not None and
+            self.get_id() and other.get_id() and
 
             # Both have same id
             self.get_id() == other.get_id()
@@ -324,15 +326,11 @@ class AbstractCrudObject(AbstractObject):
 
     @classmethod
     def get_by_ids(cls, ids, params=None, fields=None):
-        params = {} if params is None else params.copy()
-        if not fields:
-            params['fields'] = cls.get_default_read_fields()
-        else:
-            params['fields'] = ','.join(fields)
-
+        params = dict(params or {})
+        cls._assign_fields_to_params(fields, params)
         params['ids'] = ','.join(map(str, ids))
         response = FacebookAdsApi.get_default_api().call(
-            FacebookAdsApi.HTTP_METHOD_GET,
+            'GET',
             ['/'],
             params=params,
         )
@@ -347,10 +345,7 @@ class AbstractCrudObject(AbstractObject):
 
     def get_id(self):
         """Returns the object's fbid if set. Else, it returns None."""
-        if self.Field.id in self:
-            return self[self.Field.id]
-        else:
-            return None
+        return self.get(self.Field.id)
 
     def get_parent_id(self):
         """Returns the object's parent's id."""
@@ -361,10 +356,7 @@ class AbstractCrudObject(AbstractObject):
         Returns the api associated with the object. If None, returns the default
         api.
         """
-        if self._api is not None:
-            return self._api
-        else:
-            return FacebookAdsApi.get_default_api()
+        return self._api or FacebookAdsApi.get_default_api()
 
     def get_id_assured(self):
         """Returns the fbid of the object.
@@ -372,10 +364,7 @@ class AbstractCrudObject(AbstractObject):
         Raises:
             FacebookBadObjectError if the object does not have an id.
         """
-        if (
-            self.Field.id not in self or
-            self[self.Field.id] is None
-        ):
+        if not self.get(self.Field.id):
             raise FacebookBadObjectError(
                 "%s object needs an id for this operation."
                 % self.__class__.__name__
@@ -389,7 +378,7 @@ class AbstractCrudObject(AbstractObject):
         Raises:
             FacebookBadObjectError if the object does not have a parent id.
         """
-        if self.get_parent_id() is None:
+        if not self.get_parent_id():
             raise FacebookBadObjectError(
                 "%s object needs a parent_id for this operation."
                 % self.__class__.__name__
@@ -404,7 +393,7 @@ class AbstractCrudObject(AbstractObject):
             FacebookBadObjectError if get_api returns None.
         """
         api = self.get_api()
-        if api is None:
+        if not api:
             raise FacebookBadObjectError(
                 "%s does not yet have an associated api object."
                 % self.__class__.__name__
@@ -423,15 +412,11 @@ class AbstractCrudObject(AbstractObject):
         Sets object's data as if it were read from the server.
         Warning: Does not log changes.
         """
-        for key in data:
-            key = str(key)
-            value = data[key]
-
-            self._data[key] = value
+        for key in map(str, data):
+            self._data[key] = data[key]
 
             # clear history due to the update
-            if key in self._changes:
-                del self._changes[key]
+            self._changes.pop(key, None)
 
         return self
 
@@ -494,12 +479,12 @@ class AbstractCrudObject(AbstractObject):
             self if not a batch call.
             the return value of batch.add if a batch call.
         """
-        if self.get_id() is not None:
+        if self.get_id():
             raise FacebookBadObjectError(
                 "This %s object was already created." % self.__class__.__name__
             )
 
-        params = {} if params is None else params.copy()
+        params = {} if not params else params.copy()
         params.update(self.export_data())
 
         if batch is not None:
@@ -507,15 +492,15 @@ class AbstractCrudObject(AbstractObject):
                 self._set_data(response.json())
                 self._clear_history()
 
-                if success is not None:
+                if success:
                     success(response)
 
             def callback_failure(response):
-                if failure is not None:
+                if failure:
                     failure(response)
 
             batch_call = batch.add(
-                FacebookAdsApi.HTTP_METHOD_POST,
+                'POST',
                 (self.get_parent_id_assured(), self.get_endpoint()),
                 params=params,
                 files=files,
@@ -525,7 +510,7 @@ class AbstractCrudObject(AbstractObject):
             return batch_call
         else:
             response = self.get_api_assured().call(
-                FacebookAdsApi.HTTP_METHOD_POST,
+                'POST',
                 (self.get_parent_id_assured(), self.get_endpoint()),
                 params=params,
                 files=files,
@@ -563,25 +548,22 @@ class AbstractCrudObject(AbstractObject):
             self if not a batch call.
             the return value of batch.add if a batch call.
         """
-        params = {} if params is None else params.copy()
-        if not fields:
-            params['fields'] = self.get_default_read_fields() or None
-        else:
-            params['fields'] = ','.join(fields)
+        params = dict(params or {})
+        self._assign_fields_to_params(fields, params)
 
         if batch is not None:
             def callback_success(response):
                 self._set_data(response.json())
 
-                if success is not None:
+                if success:
                     success(response)
 
             def callback_failure(response):
-                if failure is not None:
+                if failure:
                     failure(response)
 
             batch_call = batch.add(
-                FacebookAdsApi.HTTP_METHOD_GET,
+                'GET',
                 self.get_node_path(),
                 params=params,
                 success=callback_success,
@@ -590,7 +572,7 @@ class AbstractCrudObject(AbstractObject):
             return batch_call
         else:
             response = self.get_api_assured().call(
-                FacebookAdsApi.HTTP_METHOD_GET,
+                'GET',
                 self.get_node_path(),
                 params=params,
             )
@@ -625,7 +607,7 @@ class AbstractCrudObject(AbstractObject):
             self if not a batch call.
             the return value of batch.add if a batch call.
         """
-        params = {} if params is None else params.copy()
+        params = {} if not params else params.copy()
         params.update(self.export_data())
         self._set_data(params)
 
@@ -633,15 +615,15 @@ class AbstractCrudObject(AbstractObject):
             def callback_success(response):
                 self._clear_history()
 
-                if success is not None:
+                if success:
                     success(response)
 
             def callback_failure(response):
-                if failure is not None:
+                if failure:
                     failure(response)
 
             batch_call = batch.add(
-                FacebookAdsApi.HTTP_METHOD_POST,
+                'POST',
                 self.get_node_path(),
                 failure=callback_failure,
                 files=files,
@@ -651,7 +633,7 @@ class AbstractCrudObject(AbstractObject):
             return batch_call
         else:
             self.get_api_assured().call(
-                FacebookAdsApi.HTTP_METHOD_POST,
+                'POST',
                 self.get_node_path(),
                 files=files,
                 params=params,
@@ -688,15 +670,15 @@ class AbstractCrudObject(AbstractObject):
             def callback_success(response):
                 self.clear_id()
 
-                if success is not None:
+                if success:
                     success(response)
 
             def callback_failure(response):
-                if failure is not None:
+                if failure:
                     failure(response)
 
             batch_call = batch.add(
-                FacebookAdsApi.HTTP_METHOD_DELETE,
+                'DELETE',
                 self.get_node_path(),
                 params=params,
                 success=callback_success,
@@ -705,7 +687,7 @@ class AbstractCrudObject(AbstractObject):
             return batch_call
         else:
             self.get_api_assured().call(
-                FacebookAdsApi.HTTP_METHOD_DELETE,
+                'DELETE',
                 self.get_node_path(),
                 params=params,
             )
@@ -720,7 +702,7 @@ class AbstractCrudObject(AbstractObject):
         Calls remote_create method if object has not been created. Else, calls
         the remote_update method.
         """
-        if self.Field.id in self:
+        if self.get_id():
             return self.remote_update(*args, **kwargs)
         else:
             return self.remote_create(*args, **kwargs)
@@ -745,7 +727,7 @@ class AbstractCrudObject(AbstractObject):
         Returns first object when iterating over EdgeIterator with argument
         self as source_object and the rest as given __init__ arguments.
         """
-        params = {} if params is None else params.copy()
+        params = {} if not params else params.copy()
         params['limit'] = '1'
         for obj in self.iterate_edge(
             target_objects_class,
@@ -899,7 +881,7 @@ class AdAccount(CannotCreate, CannotDelete, AbstractCrudObject):
 
         return my_account
 
-    def opt_out_user_from_targeting(self, schema, users, app_ids=[]):
+    def opt_out_user_from_targeting(self, schema, users, app_ids=None):
         """Opts out users from being targeted by this ad account.
 
         Args:
@@ -910,7 +892,7 @@ class AdAccount(CannotCreate, CannotDelete, AbstractCrudObject):
             Return FacebookResponse object
         """
         return self.get_api_assured().call(
-            FacebookAdsApi.HTTP_METHOD_DELETE,
+            'DELETE',
             (self.get_id_assured(), 'usersofanyaudience'),
             params=CustomAudience.format_params(schema, users, app_ids),
         )
@@ -1103,7 +1085,8 @@ class AdAccountGroupUser(AbstractCrudObject):
         return AdUser(fbid=self[self.Field.uid])
 
 
-class AdCampaign(HasStatus, HasObjective, CanArchive, AbstractCrudObject):
+class AdCampaign(CanValidate, HasStatus, HasObjective, CanArchive,
+                 AbstractCrudObject):
 
     class Field(object):
         account_id = 'account_id'
@@ -1112,6 +1095,7 @@ class AdCampaign(HasStatus, HasObjective, CanArchive, AbstractCrudObject):
         is_completed = 'is_completed'
         name = 'name'
         objective = 'objective'
+        promoted_object = 'promoted_object'
         status = 'campaign_group_status'
 
     class BuyingType(object):
@@ -1136,7 +1120,7 @@ class AdCampaign(HasStatus, HasObjective, CanArchive, AbstractCrudObject):
         return self.iterate_edge(AdStats, fields, params)
 
 
-class AdSet(HasStatus, CanArchive, AbstractCrudObject):
+class AdSet(CanValidate, HasStatus, CanArchive, AbstractCrudObject):
 
     class Field(HasBidInfo, object):
         account_id = 'account_id'
@@ -1153,6 +1137,7 @@ class AdSet(HasStatus, CanArchive, AbstractCrudObject):
         name = 'name'
         pacing_type = 'pacing_type'
         promoted_object = 'promoted_object'
+        rf_prediction_id = 'rf_prediction_id'
         start_time = 'start_time'
         status = 'campaign_status'
         targeting = 'targeting'
@@ -1227,7 +1212,7 @@ class AdGroup(HasStatus, HasObjective, CanArchive, AbstractCrudObject):
 
     def get_keyword_stats(self, fields=None, params=None):
         """Returns iterator over KeywordStats's associated with this ad."""
-        return self.iterate_edge(KeywordStats, fields, params)
+        return self.edge_object(KeywordStats, fields, params)
 
     def get_ad_preview(self, fields=None, params=None):
         """Returns AdGroupPreview object associated with this ad."""
@@ -1290,6 +1275,7 @@ class AdCreative(AbstractCrudObject):
         object_type = 'object_type'
         object_url = 'object_url'
         preview_url = 'preview_url'
+        product_set_id = 'product_set_id'
         thumbnail_url = 'thumbnail_url'
         title = 'title'
         url_tags = 'url_tags'
@@ -1321,24 +1307,60 @@ class AdImage(CannotUpdate, AbstractCrudObject):
         return (self.get_parent_id_assured(), self.get_endpoint())
 
     def _set_data(self, data):
-        data = list(data['images'].values())[0]
+        """
+            `data` may have a different structure depending if you're creating
+            new AdImages or iterating over existing ones using something like
+            AdAccount.get_ad_images().
 
-        for key in data:
-            key = str(key)
-            value = data[key]
+            While reading existing images, _set_data from AbstractCrudObject
+            handles everything correctly, but we need to treat the remote_create
+            case.
 
-            self._data[key] = value
+            remote_create sample response:
+            {
+              "images": {
+                "8cf726a44ab7008c5cc6b4ebd2491234": {
+                  "hash":"8cf726a44ab7008c5cc6b4ebd2491234",
+                  "url":"https://fbcdn-photos-a.akamaihd.net/..."
+                }
+              }
+            }
 
-            # clear history due to the update
-            if key in self._changes:
-                del self._changes[key]
+            Sample response when calling act_<ACT_ID>/adimages, used internally
+            by AdAccount.get_ad_images():
+            {
+              "data": [
+                {
+                  "hash": "181b88e3cdf6464af7ed52fe488fe559",
+                  "id": "1739564149602806:181b88e3cdf6464af7ed52fe488fe559"
+                }
+              ],
+              "paging": {
+                "cursors": {
+                  "before": "MTczOTU2NDE0OTYwMjgwNjoxODFiODh==",
+                  "after": "MTczOTU2NDE0OTYwMjgwNjoxODFiODhl=="
+                }
+              }
+            }
+        """
 
-        self._data[self.Field.id] = '%s:%s' % (
-            self.get_parent_id_assured()[4:],
-            self[self.Field.hash],
-        )
+        if 'images' in data:
+            _, data = data['images'].popitem()
 
-        return self
+            for key in map(str, data):
+                self._data[key] = data[key]
+
+                # clear history due to the update
+                self._changes.pop(key, None)
+
+            self._data[self.Field.id] = '%s:%s' % (
+                self.get_parent_id_assured()[4:],
+                self[self.Field.hash],
+            )
+
+            return self
+        else:
+            return super(AdImage, self)._set_data(data)
 
     def remote_create(
         self,
@@ -1354,7 +1376,7 @@ class AdImage(CannotUpdate, AbstractCrudObject):
         not have the files argument but requires the 'filename' property to be
         defined.
         """
-        if self[self.Field.filename] is None:
+        if not self[self.Field.filename]:
             raise FacebookBadObjectError(
                 "AdImage required a filename to be defined."
             )
@@ -1374,6 +1396,27 @@ class AdImage(CannotUpdate, AbstractCrudObject):
         """Returns the image hash to which AdCreative's can refer."""
         return self[self.Field.hash]
 
+    def remote_read(
+        self,
+        batch=None,
+        failure=None,
+        fields=None,
+        params=None,
+        success=None,
+    ):
+        if self[AdImage.Field.id]:
+            _, image_hash = self[AdImage.Field.id].split(':')
+            account = AdAccount(fbid=self.get_parent_id_assured())
+            params = {
+                'hashes': [
+                    image_hash,
+                ],
+            }
+            images = account.get_ad_images(fields=fields, params=params)
+            if images:
+                self._set_data(images[0]._data)
+
+
 
 class AdPreview(AbstractObject):
 
@@ -1408,7 +1451,9 @@ class AdCreativePreview(AdPreview):
 
 class AdGroupPreview(AdCreativePreview):
 
-    pass
+    @classmethod
+    def get_endpoint(cls):
+        return 'previews'
 
 
 # Stats for an object - e.g. {adgroup id}/stats
@@ -1521,9 +1566,9 @@ class CustomAudience(AbstractCrudObject):
         return 'customaudiences'
 
     @classmethod
-    def format_params(cls, schema, users, app_ids=[]):
+    def format_params(cls, schema, users, app_ids=None):
         hashed_users = []
-        if schema == cls.Schema.phone_hash or schema == cls.Schema.email_hash:
+        if schema in (cls.Schema.phone_hash, cls.Schema.email_hash):
             for user in users:
                 if schema == cls.Schema.email_hash:
                     user = user.strip(" \t\r\n\0\x0B.").lower()
@@ -1548,7 +1593,7 @@ class CustomAudience(AbstractCrudObject):
             'payload': payload,
         }
 
-    def add_users(self, schema, users, app_ids=[]):
+    def add_users(self, schema, users, app_ids=None):
         """Adds users to this CustomAudience.
 
         Args:
@@ -1560,12 +1605,12 @@ class CustomAudience(AbstractCrudObject):
             The FacebookResponse object.
         """
         return self.get_api_assured().call(
-            FacebookAdsApi.HTTP_METHOD_POST,
+            'POST',
             (self.get_id_assured(), 'users'),
             params=CustomAudience.format_params(schema, users, app_ids),
         )
 
-    def remove_users(self, schema, users, app_ids=[]):
+    def remove_users(self, schema, users, app_ids=None):
         """Deletes users from this CustomAudience.
 
         Args:
@@ -1577,7 +1622,7 @@ class CustomAudience(AbstractCrudObject):
             The FacebookResponse object.
         """
         return self.get_api_assured().call(
-            FacebookAdsApi.HTTP_METHOD_DELETE,
+            'DELETE',
             (self.get_id_assured(), 'users'),
             params=CustomAudience.format_params(schema, users, app_ids),
         )
@@ -1592,7 +1637,7 @@ class CustomAudience(AbstractCrudObject):
             The FacebookResponse object.
         """
         return self.get_api_assured().call(
-            FacebookAdsApi.HTTP_METHOD_POST,
+            'POST',
             (self.get_id_assured(), 'adaccounts'),
             params={'adaccounts': account_ids},
         )
@@ -1607,7 +1652,7 @@ class CustomAudience(AbstractCrudObject):
             The FacebookResponse object.
         """
         return self.get_api_assured().call(
-            FacebookAdsApi.HTTP_METHOD_DELETE,
+            'DELETE',
             (self.get_id_assured(), 'users'),
             params={'adaccounts': account_ids},
         )
@@ -1775,7 +1820,7 @@ class ReachFrequencyPrediction(AbstractCrudObject):
         params = dict((k, v) for k, v in params.items() if v is not None)
 
         response = self.get_api_assured().call(
-            FacebookAdsApi.HTTP_METHOD_POST,
+            'POST',
             (self.get_parent_id_assured(), self.get_endpoint()),
             params=params,
         )
@@ -1788,7 +1833,7 @@ class ReachFrequencyPrediction(AbstractCrudObject):
             self.Field.action: self.Action.cancel,
         }
         self.get_api_assured().call(
-            FacebookAdsApi.HTTP_METHOD_POST,
+            'POST',
             (self.get_parent_id_assured(), self.get_endpoint()),
             params=params,
         )
@@ -1855,6 +1900,7 @@ class TargetingSpecsField(object):
     conjunctive_user_adclusters = 'conjunctive_user_adclusters'
     connections = 'connections'
     custom_audiences = 'custom_audiences'
+    dynamic_audience_ids = 'dynamic_audience_ids'
     education_majors = 'education_majors'
     education_schools = 'education_schools'
     education_statuses = 'education_statuses'
@@ -1923,3 +1969,275 @@ class AutoComplete(AbstractCrudObject):
         return (
             self.get_endpoint(),
         )
+
+
+class Business(AbstractCrudObject, CannotCreate, CannotDelete):
+
+    class Field(object):
+        created_by = 'created_by'
+        creation_time = 'creation_time'
+        id = 'id'
+        name = 'name'
+        primary_page = 'primary_page'
+        timezone_id = 'timezone_id'
+        update_time = 'update_time'
+        updated_by = 'updated_by'
+        vertical_id = 'vertical_id'
+
+    def get_product_catalogs(self, fields=None, params=None):
+        return self.iterate_edge(ProductCatalog, fields, params)
+
+
+class ProductCatalog(AbstractCrudObject):
+
+    class Field(object):
+        id = 'id'
+        name = 'name'
+
+    class Role(object):
+        admin = 'ADMIN'
+
+    class Availability(object):
+        available_for_order = 'available for order'
+        in_stock = 'in stock'
+        out_of_stock = 'out of stock'
+        preorder = 'preorder'
+
+    class AgeGroup(object):
+        adult = 'adult'
+        infant = 'infant'
+        kids = 'kids'
+        newborn = 'newborn'
+        toddler = 'toddler'
+
+    class Gender(object):
+        men = 'men'
+        women = 'women'
+        unisex = 'unisex'
+
+    class Condition(object):
+        new = 'new'
+        refurbished = 'refurbished'
+        used = 'used'
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'product_catalogs'
+
+    def get_product_feeds(self, fields=None, params=None):
+        return self.iterate_edge(ProductFeed, fields, params)
+
+    def add_users(self, user, role):
+        params = {
+            user: user,
+            role: role,
+        }
+        return self.get_api_assured().call(
+            'POST',
+            (self.get_id_assured(), 'userpermissions'),
+            params=params
+        )
+
+    def remove_users(self, user):
+        params = {
+            user: user,
+        }
+        return self.get_api_assured().call(
+            'DELETE',
+            (self.get_id_assured(), 'userpermissions'),
+            params=params
+        )
+
+    def add_external_event_sources(self, pixel_ids):
+        params = {
+            'external_event_sources': pixel_ids,
+        }
+        return self.get_api_assured().call(
+            'POST',
+            (self.get_id_assured(), 'external_event_sources'),
+            params=params
+        )
+
+    def remove_external_event_sources(self, pixel_ids):
+        params = {
+            'external_event_sources': pixel_ids,
+        }
+        return self.get_api_assured().call(
+            'DELETE',
+            (self.get_id_assured(), 'external_event_sources'),
+            params=params
+        )
+
+    def get_external_event_sources(self, fields=None, params=None):
+        return self.iterate_edge(
+            ProductCatalogExternalEventSource,
+            fields,
+            params
+        )
+
+
+class ProductCatalogExternalEventSource(
+    CannotCreate,
+    CannotDelete,
+    CannotUpdate,
+    AbstractCrudObject
+):
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'external_event_sources'
+
+
+class ProductFeed(AbstractCrudObject):
+
+    class Field(object):
+        country = 'country'
+        delimiter = 'delimiter'
+        encoding = 'encoding'
+        file_name = 'file_name'
+        format = 'format'
+        id = 'id'
+        name = 'name'
+        quotes = 'quotes'
+        schedule = 'schedule'
+
+    class Format(object):
+        tsv = 'TSV'
+        xml = 'XML'
+
+    class Encoding(object):
+        ascii = 'ASCII'
+        autodetect = 'AUTODETECT'
+        latin_1 = 'LATIN_1'
+        utf_16 = 'UTF_16'
+        utf_8 = 'UTF_8'
+
+    class Delimiter(object):
+        pipe = 'PIPE'
+        tab = 'TAB'
+        tilde = 'TILDE'
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'product_feeds'
+
+    def get_products(self, fields=None, params=None):
+        return self.iterate_edge(Product, fields, params)
+
+
+class ProductFeedUpload(AbstractCrudObject):
+
+    class Field(object):
+        end_time = 'end_time'
+        id = 'id'
+        start_time = 'start_time'
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'uploads'
+
+    def get_errors(self, fields=None, params=None):
+        return self.iterate_edge(ProductFeedUploadError, fields, params)
+
+
+class ProductFeedUploadError(AbstractCrudObject):
+
+    class Field(object):
+        column_number = 'column_number'
+        description = 'description'
+        id = 'id'
+        row_number = 'row_number'
+        severity = 'severity'
+        summary = 'summary'
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'errors'
+
+
+class ProductSet(AbstractCrudObject):
+
+    class Field(object):
+        filter = 'filter'
+        id = 'id'
+        name = 'name'
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'product_sets'
+
+    def get_product_groups(self, fields=None, params=None):
+        return self.iterate_edge(ProductGroup, fields, params)
+
+    def get_products(self, fields=None, params=None):
+        return self.iterate_edge(Product, fields, params)
+
+
+class ProductGroup(AbstractCrudObject):
+
+    class Field(object):
+        id = 'id'
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'product_groups'
+
+
+class Product(AbstractCrudObject):
+
+    class Field(object):
+        additional_image_link = 'additional_image_link'
+        age_group = 'age_group'
+        availability = 'availability'
+        brand = 'brand'
+        color = 'color'
+        condition = 'condition'
+        description = 'description'
+        expiration_date = 'expiration_date'
+        gender = 'gender'
+        google_product_category = 'google_product_category'
+        gtin = 'gtin'
+        id = 'id'
+        image_link = 'image_link'
+        item_group_id = 'item_group_id'
+        link = 'link'
+        material = 'material'
+        mpn = 'mpn'
+        pattern = 'pattern'
+        price = 'price'
+        product_type = 'product_type'
+        sale_price = 'sale_price'
+        sale_price_effective_date = 'sale_price_effective_date'
+        shipping = 'shipping'
+        shipping_size = 'shipping_size'
+        shipping_weight = 'shipping_weight'
+        title = 'title'
+
+    class Availability(object):
+        in_stock = 'IN_STOCK'
+        out_of_stock = 'OUT_OF_STOCK'
+        preorder = 'PREORDER'
+
+    class Condition(object):
+        new = 'PC_NEW'
+        refurbished = 'PC_REFURBISHED'
+        used = 'PC_USED'
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'products'
+
+
+class ProductAudience(CannotUpdate, CannotDelete, AbstractCrudObject):
+
+    class Field(object):
+        description = 'description'
+        exclusions = 'exclusions'
+        inclusions = 'inclusions'
+        name = 'name'
+        pixel_id = 'pixel_id'
+        product_set_id = 'product_set_id'
+
+    @classmethod
+    def get_endpoint(cls):
+        return 'product_audiences'
